/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.factories;

import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.ConfigurationException;
import org.jboss.cache.factories.annotations.DefaultFactoryFor;
import org.jboss.cache.interceptors.*;

import java.util.ArrayList;
import java.util.List;

/**
 * Factory class that builds an interceptor chain based on cache configuration.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
@DefaultFactoryFor(classes = Interceptor.class)
public class InterceptorChainFactory extends ComponentFactory
{
   /**
    * Note - this method used to return a singleton instance, and since 2.1.0 returns a new instance.  The method is
    * deprecated and you should use the no-arg constructor to create a new instance of this factory.
    *
    * @return a NEW instance of this class.
    */
   @Deprecated
   public static InterceptorChainFactory getInstance()
   {
      return new InterceptorChainFactory();
   }

   public Interceptor buildInterceptorChain() throws IllegalAccessException, ClassNotFoundException, InstantiationException
   {
      if (configuration.isNodeLockingOptimistic())
      {
         return createOptimisticInterceptorChain();
      }
      else
      {
         return createPessimisticInterceptorChain();
      }
   }

   public static Interceptor setLastInterceptorPointer(Interceptor first, Interceptor last)
   {
      Interceptor i = first;
      while (i != null)
      {
         i.setLast(last);
         i = i.getNext();
      }
      return first;
   }

   private Interceptor createInterceptor(Class<? extends Interceptor> clazz) throws IllegalAccessException, InstantiationException
   {
      Interceptor i = componentRegistry.getComponent(clazz.getName(), clazz);
      if (i == null)
      {
         i = clazz.newInstance();
//         componentRegistry.wireDependencies(i);
         // add this interceptor as a NAMED component into the registry
         componentRegistry.registerComponent(clazz.getName(), i, clazz);
      }
      else
      {
         // wipe next/last chaining!!
         i.setLast(null);
         i.setNext(null);
      }
      return i;
   }

   /**
    * Adds an interceptor at the end of the chain
    */
   private void addInterceptor(Interceptor first, Interceptor i)
   {
      if (first == null) return;
      while (first.getNext() != null) first = first.getNext();
      first.setNext(i);
   }

   private Interceptor createPessimisticInterceptorChain() throws IllegalAccessException, InstantiationException, ClassNotFoundException
   {
      Interceptor call_interceptor;
      Interceptor lock_interceptor;
      Interceptor repl_interceptor = null;
      Interceptor cache_loader_interceptor = null;
      Interceptor cache_store_interceptor = null;
      Interceptor unlock_interceptor;
      Interceptor passivation_interceptor = null;
      Interceptor activation_interceptor = null;
      Interceptor cacheMgmtInterceptor;
      Interceptor txInterceptor;
      Interceptor eviction_interceptor;
      Interceptor dataGravitatorInterceptor = null;
      Interceptor invocationCtxInterceptor = createInterceptor(InvocationContextInterceptor.class);
      Interceptor notificationInterceptor = createInterceptor(NotificationInterceptor.class);
      Interceptor first = null;


      call_interceptor = createInterceptor(CallInterceptor.class);

      if (isUsingBuddyReplication()) dataGravitatorInterceptor = createInterceptor(DataGravitatorInterceptor.class);

      lock_interceptor = createInterceptor(PessimisticLockInterceptor.class);

      unlock_interceptor = createInterceptor(UnlockInterceptor.class);

      cacheMgmtInterceptor = createInterceptor(CacheMgmtInterceptor.class);

      txInterceptor = createInterceptor(TxInterceptor.class);

      switch (configuration.getCacheMode())
      {
         case REPL_SYNC:
         case REPL_ASYNC:
            repl_interceptor = createInterceptor(ReplicationInterceptor.class);
            break;
         case INVALIDATION_SYNC:
         case INVALIDATION_ASYNC:
            repl_interceptor = createInterceptor(InvalidationInterceptor.class);
            break;
         case LOCAL:
            //Nothing...
      }

      if (isUsingCacheLoaders())
      {
         if (configuration.getCacheLoaderConfig().isPassivation())
         {
            activation_interceptor = createInterceptor(ActivationInterceptor.class);
            passivation_interceptor = createInterceptor(PassivationInterceptor.class);
         }
         else
         {
            cache_loader_interceptor = createInterceptor(CacheLoaderInterceptor.class);
            cache_store_interceptor = createInterceptor(CacheStoreInterceptor.class);
         }
      }

      // load the icInterceptor first
      if (first == null) first = invocationCtxInterceptor;

      // load the cache management interceptor next
      if (configuration.getExposeManagementStatistics())
      {
         if (first == null)
         {
            first = cacheMgmtInterceptor;
         }
         else
         {
            addInterceptor(first, cacheMgmtInterceptor);
         }
      }

      // load the tx interceptor
      if (first == null)
      {
         first = txInterceptor;
      }
      else
      {
         addInterceptor(first, txInterceptor);
      }

      if (first == null)
         first = notificationInterceptor;
      else
         addInterceptor(first, notificationInterceptor);

      if (repl_interceptor != null)
      {
         if (first == null)
         {
            first = repl_interceptor;
         }
         else
         {
            addInterceptor(first, repl_interceptor);
         }
      }

      if (first == null)
      {
         first = lock_interceptor;
      }
      else
      {
         addInterceptor(first, lock_interceptor);
      }


      if (unlock_interceptor != null)
      {
         if (first == null)
         {
            first = unlock_interceptor;
         }
         else
         {
            addInterceptor(first, unlock_interceptor);
         }
      }

      if (activation_interceptor != null)
      {
         if (first == null)
         {
            first = activation_interceptor;
         }
         else
         {
            addInterceptor(first, activation_interceptor);
         }
         if (first == null)
         {
            first = passivation_interceptor;
         }
         else
         {
            addInterceptor(first, passivation_interceptor);
         }
      }

      if (cache_loader_interceptor != null)
      {
         if (first == null)
         {
            first = cache_loader_interceptor;
         }
         else
         {
            addInterceptor(first, cache_loader_interceptor);
         }
         if (first == null)
         {
            first = cache_store_interceptor;
         }
         else
         {
            addInterceptor(first, cache_store_interceptor);
         }
      }

      if (dataGravitatorInterceptor != null)
      {
         if (first == null)
         {
            first = dataGravitatorInterceptor;
         }
         else
         {
            addInterceptor(first, dataGravitatorInterceptor);
         }
      }

      if (configuration.getEvictionConfig() != null && configuration.getEvictionConfig().isValidConfig())
      {
         eviction_interceptor = createInterceptor(EvictionInterceptor.class);
         if (first == null)
         {
            first = eviction_interceptor;
         }
         else
         {
            addInterceptor(first, eviction_interceptor);
         }
      }

      if (first == null)
      {
         first = call_interceptor;
      }
      else
      {
         addInterceptor(first, call_interceptor);
      }

      return setLastInterceptorPointer(first, call_interceptor);
   }

   private boolean isUsingCacheLoaders()
   {
      return configuration.getCacheLoaderConfig() != null && !configuration.getCacheLoaderConfig().getIndividualCacheLoaderConfigs().isEmpty();
   }

   private boolean isUsingBuddyReplication()
   {
      return configuration.getBuddyReplicationConfig() != null && configuration.getBuddyReplicationConfig().isEnabled();
   }

   private Interceptor createOptimisticInterceptorChain() throws IllegalAccessException, InstantiationException, ClassNotFoundException
   {
      Interceptor txInterceptor, replicationInterceptor = null, lockInterceptor, validationInterceptor;
      Interceptor createIfNotExistsInterceptor, nodeInterceptor, invokerInterceptor, activationInterceptor = null;
      Interceptor passivationInterceptor = null, cacheLoaderInterceptor = null, cacheStoreInterceptor = null, first = null;
      Interceptor cacheMgmtInterceptor, evictionInterceptor = null, dataGravitatorInterceptor = null;
      Interceptor invocationCtxInterceptor = createInterceptor(InvocationContextInterceptor.class);
      Interceptor notificationInterceptor = createInterceptor(NotificationInterceptor.class);

      if (isUsingCacheLoaders())
      {
         if (configuration.getCacheLoaderConfig().isPassivation())
         {
            activationInterceptor = createInterceptor(ActivationInterceptor.class);
            passivationInterceptor = createInterceptor(PassivationInterceptor.class);
         }
         else
         {
            cacheLoaderInterceptor = createInterceptor(CacheLoaderInterceptor.class);
            cacheStoreInterceptor = createInterceptor(CacheStoreInterceptor.class);
         }
      }

      txInterceptor = createInterceptor(TxInterceptor.class);

      if (isUsingBuddyReplication()) dataGravitatorInterceptor = createInterceptor(DataGravitatorInterceptor.class);

      switch (configuration.getCacheMode())
      {
         case REPL_SYNC:
         case REPL_ASYNC:
            replicationInterceptor = createInterceptor(OptimisticReplicationInterceptor.class);
            break;
         case INVALIDATION_SYNC:
         case INVALIDATION_ASYNC:
            replicationInterceptor = createInterceptor(InvalidationInterceptor.class);
            break;
         case LOCAL:
            //Nothing...
      }

      lockInterceptor = createInterceptor(OptimisticLockingInterceptor.class);

      validationInterceptor = createInterceptor(OptimisticValidatorInterceptor.class);

      createIfNotExistsInterceptor = createInterceptor(OptimisticCreateIfNotExistsInterceptor.class);

      nodeInterceptor = createInterceptor(OptimisticNodeInterceptor.class);

      invokerInterceptor = createInterceptor(CallInterceptor.class);

      if (configuration.getEvictionConfig() != null && configuration.getEvictionConfig().isValidConfig())
      {
         evictionInterceptor = createInterceptor(EvictionInterceptor.class);
      }

      if (first == null) first = invocationCtxInterceptor;

      if (configuration.getExposeManagementStatistics())
      {
         cacheMgmtInterceptor = createInterceptor(CacheMgmtInterceptor.class);
         if (first == null)
         {
            first = cacheMgmtInterceptor;
         }
         else
         {
            addInterceptor(first, cacheMgmtInterceptor);
         }
      }

      if (txInterceptor != null)
      {
         if (first == null)
         {
            first = txInterceptor;
         }
         else
         {
            addInterceptor(first, txInterceptor);
         }
      }

      if (first == null)
         first = notificationInterceptor;
      else
         addInterceptor(first, notificationInterceptor);

      if (first == null)
      {
         first = replicationInterceptor;
      }
      else
      {
         addInterceptor(first, replicationInterceptor);
      }

      if (passivationInterceptor != null && !configuration.getCacheLoaderConfig().isFetchPersistentState())
      {
         if (first == null)
         {
            first = passivationInterceptor;
         }
         else
         {
            addInterceptor(first, passivationInterceptor);
         }
      }

      // add the cache store interceptor here
      if (cacheStoreInterceptor != null && !configuration.getCacheLoaderConfig().isFetchPersistentState())
      {
         if (first == null)
         {
            first = cacheStoreInterceptor;
         }
         else
         {
            addInterceptor(first, cacheStoreInterceptor);
         }
      }

      // cache loader interceptor is only invoked if we are ready to write to the actual tree cache
      if (activationInterceptor != null)
      {
         if (first == null)
         {
            first = activationInterceptor;
         }
         else
         {
            addInterceptor(first, activationInterceptor);
         }

         if (configuration.getCacheLoaderConfig().isFetchPersistentState())
         {
            if (first == null)
            {
               first = passivationInterceptor;
            }
            else
            {
               addInterceptor(first, passivationInterceptor);
            }
         }
      }

      if (cacheLoaderInterceptor != null)
      {
         if (first == null)
         {
            first = cacheLoaderInterceptor;
         }
         else
         {
            addInterceptor(first, cacheLoaderInterceptor);
         }

         if (configuration.getCacheLoaderConfig().isFetchPersistentState())
         {
            if (first == null)
            {
               first = cacheStoreInterceptor;
            }
            else
            {
               addInterceptor(first, cacheStoreInterceptor);
            }
         }
      }

      if (dataGravitatorInterceptor != null)
      {
         if (first == null)
         {
            first = dataGravitatorInterceptor;
         }
         else
         {
            addInterceptor(first, dataGravitatorInterceptor);
         }
      }


      if (first == null)
      {
         first = lockInterceptor;
      }
      else
      {
         addInterceptor(first, lockInterceptor);
      }

      if (first == null)
      {
         first = validationInterceptor;
      }
      else
      {
         addInterceptor(first, validationInterceptor);
      }

      if (first == null)
      {
         first = createIfNotExistsInterceptor;
      }
      else
      {
         addInterceptor(first, createIfNotExistsInterceptor);
      }

      // eviction interceptor to come before the optimistic node interceptor
      if (first == null)
      {
         first = evictionInterceptor;
      }
      else
      {
         addInterceptor(first, evictionInterceptor);
      }

      if (first == null)
      {
         first = nodeInterceptor;
      }
      else
      {
         addInterceptor(first, nodeInterceptor);
      }


      if (first == null)
      {
         first = invokerInterceptor;
      }
      else
      {
         addInterceptor(first, invokerInterceptor);
      }

      return setLastInterceptorPointer(first, invokerInterceptor);
   }

   public static List<Interceptor> asList(Interceptor interceptor)
   {
      if (interceptor == null)
      {
         return null;
      }
      int num = 1;
      Interceptor tmp = interceptor;
      while ((tmp = tmp.getNext()) != null)
      {
         num++;
      }
      List<Interceptor> retval = new ArrayList<Interceptor>(num);
      tmp = interceptor;
      num = 0;
      do
      {
         retval.add(tmp);
         tmp = tmp.getNext();
      }
      while (tmp != null);
      return retval;
   }

   /**
    * The same as {@link #correctInterceptorChaining(java.util.List)}
    * except that it takes in the first interceptor in the chain instead of a list of interceptors, and traverses the chain to
    * generate a list.  Identical to calling
    * <pre>
    *    InterceptorChainFactory.correctInterceptorChaining( InterceptorChainFactory.asList(first), configuration, componentRegistry );
    * </pre>
    *
    * @param first first interceptor in the chain to correct.
    * @return the first interceptor in the chain.
    */
   public Interceptor correctInterceptorChaining(Interceptor first)
   {
      return correctInterceptorChaining(asList(first));
   }

   /**
    * "Fixes" the next() and last() pointers for each interceptor, based on the order presented in the list passed in, and
    * also re-assesses dependencies for each interceptor, injecting dependencies accordingingly.
    *
    * @param interceptors interceptor chain to correct
    * @return the first interceptor in the chain.
    */
   public Interceptor correctInterceptorChaining(List<Interceptor> interceptors)
   {

      Interceptor first = null, last = null;

      for (Interceptor next : interceptors)
      {
         if (first == null)
         {
            first = last = next;
            continue;
         }
         last.setNext(next);
         last = next;
      }

      if (last != null) last.setNext(null);

      // now set the 'last' pointer.
      Interceptor i = setLastInterceptorPointer(first, last);
      // re-register this interceptor
      componentRegistry.registerComponent(Interceptor.class.getName(), i, Interceptor.class);
      return i;
   }

   /**
    * Initialises the interceptors with a CacheSPI instance.  Essentially calls setCache().  This is done AFTER the interceptor chain
    * is created, and when Cache.start() is called, so the interceptors get a fully initialised cache.
    *
    * @param componentName
    */
//   public void initialiseInterceptors(Interceptor firstInterceptor, CacheSPI cache)
//   {
//      for (Interceptor i : asList(firstInterceptor)) i.setCache(cache);
//   }
   @Override
   @SuppressWarnings("unchecked")
   protected <T> T construct(String componentName, Class<T> componentType)
   {
      try
      {
         return (T) buildInterceptorChain();
      }
      catch (Exception e)
      {
         throw new ConfigurationException("Unable to build interceptor chain", e);
      }
   }

   public static InterceptorChainFactory getInstance(ComponentRegistry componentRegistry, Configuration configuration)
   {
      InterceptorChainFactory icf = new InterceptorChainFactory();
      icf.componentRegistry = componentRegistry;
      icf.configuration = configuration;
      return icf;
   }
}
