/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.config.Option;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.optimistic.DefaultDataVersion;
import org.jboss.cache.optimistic.TransactionWorkspace;
import org.jboss.cache.optimistic.WorkspaceNode;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.transaction.TransactionEntry;
import org.jboss.cache.transaction.TransactionTable;
import org.jgroups.Address;

import javax.transaction.SystemException;
import javax.transaction.Transaction;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * This interceptor acts as a replacement to the replication interceptor when
 * the CacheImpl is configured with ClusteredSyncMode as INVALIDATE.
 * <p/>
 * The idea is that rather than replicating changes to all caches in a cluster
 * when CRUD (Create, Remove, Update, Delete) methods are called, simply call
 * evict(Fqn) on the remote caches for each changed node.  This allows the
 * remote node to look up the value in a shared cache loader which would have
 * been updated with the changes.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class InvalidationInterceptor extends BaseRpcInterceptor implements InvalidationInterceptorMBean
{
   private long m_invalidations = 0;
   protected TransactionTable txTable;
   protected Map<GlobalTransaction, List<MethodCall>> txMods;
   protected boolean optimistic;

   public InvalidationInterceptor()
   {
      initLogger();
   }

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      optimistic = configuration.isNodeLockingOptimistic();
      if (optimistic) txMods = new ConcurrentHashMap<GlobalTransaction, List<MethodCall>>();
   }

   @Inject
   private void injectDependencies(TransactionTable txTable)
   {
      this.txTable = txTable;
   }

   protected boolean skipMethodCall(InvocationContext ctx)
   {
      Option optionOverride = ctx.getOptionOverrides();
      if (optionOverride != null && optionOverride.isCacheModeLocal() && (ctx.getTransaction() == null ||
            MethodDeclarations.isTransactionLifecycleMethod(ctx.getMethodCall().getMethodId())))
      {
         // skip replication!!
         return true;
      }
      if (trace) log.trace("(" + cache.getLocalAddress() + ") method call " + ctx.getMethodCall());
      return false;
   }

   protected Object handlePutDataMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Map data, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handlePutDataEraseMethod(InvocationContext ctx, GlobalTransaction gt, Fqn fqn, Map newData, boolean createUndoOps, boolean eraseContents) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handlePutKeyValueMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handlePutDataEraseVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Map data, boolean createUndoOps, boolean eraseContent, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handlePutDataVersionedMethod(InvocationContext ctx, GlobalTransaction globalTransaction, Fqn fqn, Map map, Boolean createUndoOps, DataVersion dataVersion) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handlePutKeyValueVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value, boolean createUndoOps, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handleRemoveNodeMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handleRemoveKeyMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Object key, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handleRemoveDataMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handleDataGravitationCleanupMethod(InvocationContext ctx, Fqn primary, Fqn backup) throws Throwable
   {
      return handleCrudMethod(ctx, primary, null);
   }

   protected Object handleMoveMethod(InvocationContext ctx, Fqn from, Fqn to) throws Throwable
   {
      return handleCrudMethod(ctx, to, from);
   }

   protected Object handleRemoveKeyVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, boolean createUndoOps, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handleRemoveNodeVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, boolean createUndoOps, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handleRemoveDataVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, boolean createUndoOps, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx, fqn, null);
   }

   protected Object handlePrepareMethod(InvocationContext ctx, GlobalTransaction gtx, List modification, Address coordinator, boolean onePhaseCommit) throws Throwable
   {
      Object retval = nextInterceptor(ctx);
      Transaction tx = ctx.getTransaction();
      if (tx != null && !optimistic)
      {
         log.debug("Entering InvalidationInterceptor's prepare phase");
         // fetch the modifications before the transaction is committed (and thus removed from the txTable)
         gtx = ctx.getGlobalTransaction();
         TransactionEntry entry = txTable.get(gtx);
         if (entry == null) throw new IllegalStateException("cannot find transaction entry for " + gtx);
         List<MethodCall> modifications = new LinkedList<MethodCall>(entry.getModifications());
         if (modifications.size() > 0)
         {
            broadcastInvalidate(modifications, gtx, tx, ctx);
         }
         else
         {
            log.debug("Nothing to invalidate - no modifications in the transaction.");
         }
      }
      return retval;
   }

   protected Object handleOptimisticPrepareMethod(InvocationContext ctx, GlobalTransaction gtx, List modifications, Map data, Address address, boolean onePhaseCommit) throws Throwable
   {
      Object retval = nextInterceptor(ctx);
      Transaction tx = ctx.getTransaction();
      if (tx != null)
      {
         // here we just record the modifications but actually do the invalidate in commit.
         gtx = ctx.getGlobalTransaction();
         TransactionEntry entry = txTable.get(gtx);
         if (entry == null) throw new IllegalStateException("cannot find transaction entry for " + gtx);
         modifications = new LinkedList<MethodCall>(entry.getModifications());
         if (modifications.size() > 0)
         {
            txMods.put(gtx, modifications);
         }
      }
      return retval;
   }

   protected Object handleCommitMethod(InvocationContext ctx, GlobalTransaction globalTransaction) throws Throwable
   {
      Object retval = nextInterceptor(ctx);
      Transaction tx = ctx.getTransaction();
      if (tx != null && optimistic)
      {
         GlobalTransaction gtx = ctx.getGlobalTransaction();
         List modifications = txMods.remove(gtx);
         broadcastInvalidate(modifications, gtx, tx, ctx);
         log.debug("Committing.  Broadcasting invalidations.");
      }
      return retval;
   }

   protected Object handleRollbackMethod(InvocationContext ctx, GlobalTransaction globalTransaction) throws Throwable
   {
      Object retval = nextInterceptor(ctx);
      Transaction tx = ctx.getTransaction();
      if (tx != null && optimistic)
      {
         GlobalTransaction gtx = ctx.getGlobalTransaction();
         txMods.remove(gtx);
         log.debug("Caught a rollback.  Clearing modification in txMods");
      }
      return retval;
   }

   /**
    * @param from is only present for move operations, else pass it in as null
    */
   private Object handleCrudMethod(InvocationContext ctx, Fqn targetFqn, Fqn from)
         throws Throwable
   {
      Object retval = nextInterceptor(ctx);
      Transaction tx = ctx.getTransaction();
      Option optionOverride = ctx.getOptionOverrides();
      if (log.isDebugEnabled()) log.debug("Is a CRUD method");
      Set<Fqn> fqns = new HashSet<Fqn>();
      if (from != null)
      {
         fqns.add(from);
      }
      fqns.add(targetFqn);
      if (!fqns.isEmpty())
      {
         // could be potentially TRANSACTIONAL.  Ignore if it is, until we see a prepare().
         if (tx == null || !isValid(tx))
         {
            // the no-tx case:
            //replicate an evict call.
            for (Fqn fqn : fqns) invalidateAcrossCluster(fqn, null, isSynchronous(optionOverride), ctx);
         }
      }
      return retval;
   }

   private void broadcastInvalidate(List<MethodCall> modifications, GlobalTransaction gtx, Transaction tx, InvocationContext ctx)
   {
      if (containsPutForExternalRead(modifications))
      {
         log.debug("Modification list contains a putForExternalRead operation.  Not invalidating.");
      }
      else
      {
         try
         {
            invalidateModifications(modifications, configuration.isNodeLockingOptimistic() ? getWorkspace(gtx) : null, defaultSynchronous, ctx);
         }
         catch (Throwable t)
         {
            log.warn("Unable to broadcast evicts as a part of the prepare phase.  Rolling back.", t);
            try
            {
               tx.setRollbackOnly();
            }
            catch (SystemException se)
            {
               throw new RuntimeException("setting tx rollback failed ", se);
            }
            if (t instanceof RuntimeException)
               throw (RuntimeException) t;
            else
               throw new RuntimeException("Unable to broadcast invalidation messages", t);
         }
      }
   }

   private boolean containsPutForExternalRead(List<MethodCall> l)
   {
      if (l == null) return false;

      for (MethodCall m : l)
         if (m.getMethodId() == MethodDeclarations.putForExternalReadMethodLocal_id || m.getMethodId() == MethodDeclarations.putForExternalReadVersionedMethodLocal_id)
            return true;
      return false;
   }

   public long getInvalidations()
   {
      return m_invalidations;
   }

   public void resetStatistics()
   {
      m_invalidations = 0;
   }

   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = new HashMap<String, Object>();
      retval.put("Invalidations", m_invalidations);
      return retval;
   }

   /**
    * Same as <code>invalidateAcrossCluster(fqn, workspace, defaultSynchronous)</code>
    *
    * @deprecated use {@link #invalidateAcrossCluster(Fqn,TransactionWorkspace,boolean,InvocationContext)
    */
   protected void invalidateAcrossCluster(Fqn fqn, TransactionWorkspace workspace, InvocationContext ctx) throws Throwable
   {
      invalidateAcrossCluster(fqn, workspace, defaultSynchronous, ctx);
   }

   protected void invalidateAcrossCluster(Fqn fqn, TransactionWorkspace workspace, boolean synchronous, InvocationContext ctx) throws Throwable
   {
      // increment invalidations counter if statistics maintained
      if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
         m_invalidations++;

      // only propagate version details if we're using explicit versioning.
      /*
      MethodCall call = workspace != null && !workspace.isVersioningImplicit() ?
              MethodCallFactory.create(MethodDeclarations.evictVersionedNodeMethodLocal, fqn, workspace.getNode(fqn).getVersion()) :
              MethodCallFactory.create(MethodDeclarations.evictNodeMethodLocal, fqn);
              */
      MethodCall call = MethodCallFactory.create(MethodDeclarations.invalidateMethodLocal_id, fqn, getNodeVersion(workspace, fqn));

      if (log.isDebugEnabled()) log.debug("Cache [" + cache.getLocalAddress() + "] replicating " + call);
      // voila, invalidated!
      replicateCall(ctx, call, synchronous, ctx.getOptionOverrides());
   }

   protected DataVersion getNodeVersion(TransactionWorkspace w, Fqn f)
   {
      if (w == null) return null;
      WorkspaceNode wn = w.getNode(f);
      DataVersion v = wn.getVersion();

      if (wn.isVersioningImplicit())
      {
         // then send back an incremented version
         v = ((DefaultDataVersion) v).increment();
      }

      return v;
   }

   /**
    * Same as <code>invalidateModifications(modifications, workspace, defaultSynchronous)</code>
    *
    * @deprecated use {@link #invalidateAcrossCluster(org.jboss.cache.Fqn,org.jboss.cache.optimistic.TransactionWorkspace,boolean,InvocationContext)}
    */
   protected void invalidateModifications(List<MethodCall> modifications, TransactionWorkspace workspace, InvocationContext ctx) throws Throwable
   {
      invalidateModifications(modifications, workspace, defaultSynchronous, ctx);
   }

   protected void invalidateModifications(List<MethodCall> modifications, TransactionWorkspace workspace, boolean synchronous, InvocationContext ctx) throws Throwable
   {
      // optimise the calls list here.
      Set<Fqn> modifiedFqns = optimisedIterator(modifications);
      for (Fqn fqn : modifiedFqns) invalidateAcrossCluster(fqn, workspace, synchronous, ctx);
   }

   protected TransactionWorkspace getWorkspace(GlobalTransaction gtx)
   {
      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) txTable.get(gtx);
      return entry.getTransactionWorkSpace();
   }

   @SuppressWarnings("unchecked")
   @Deprecated
   protected void findAndAddFqns(Object[] objects, Set<Fqn> fqns, boolean isMove)
   {
      if (isMove)
      {
         Fqn f = (Fqn) objects[0];
         fqns.add(f);
         // now if this is a "move" operation, then we also have another Fqn -
         Object le = f.getLastElement();
         Fqn parent = (Fqn) objects[1];
         fqns.add(new Fqn(parent, le));
      }
      else
      {
         // it *should* be the 2nd param...
         fqns.add((Fqn) objects[1]);
      }
   }

   /**
    * Removes non-crud methods, plus clobs together common calls to Fqn's.
    * E.g, if we have put("/a/b", "1", "2") followed by a put("/a/b", "3",
    * "4") we should only evict "/a/b" once.
    *
    * @param list
    * @return Iterator containing a unique set of Fqns of crud methods in this tx
    */
   protected Set<Fqn> optimisedIterator(List<MethodCall> list)
   {
      Set<Fqn> fqns = new HashSet<Fqn>();
      if (list != null)
      {
         for (MethodCall mc : list)
         {
            if (MethodDeclarations.isCrudMethod(mc.getMethodId()))
            {
               findAndAddFqns(mc.getArgs(), fqns, MethodDeclarations.moveMethodLocal_id == mc.getMethodId());
            }
         }
      }
      return fqns;
   }
}
