/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.marshall;

import org.jboss.cache.Fqn;
import org.jboss.cache.RegionManager;
import org.jboss.cache.config.Configuration;
import static org.testng.AssertJUnit.*;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.List;

@Test(groups = "functional")
public abstract class CacheMarshallerTestBase extends AbstractVersionAwareMarshallerTest
{
   protected String currentVersion;
   protected int currentVersionShort;
   protected Class expectedMarshallerClass, latestMarshallerClass = CacheMarshaller210.class;
   protected VersionAwareMarshaller marshaller;
   protected RegionManager regionManager;
   protected Configuration c;

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      regionManager = new RegionManager();
      c = new Configuration();
      c.setUseRegionBasedMarshalling(false);
      c.setInactiveOnStartup(false);
      c.setReplVersionString(currentVersion);
      marshaller = createVAM(c, regionManager);
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown()
   {
      marshaller = null;
   }

   protected void assertMethodCallsEquals(MethodCall call1, MethodCall call2)
   {
      if (call1 == call2) return;

      assertEquals("Method IDs should match", call1.getMethodId(), call2.getMethodId());
      assertEquals("Method names should match", call1.getName(), call2.getName());
      assertEquals("Method reflection objects should match", call1.getMethod(), call2.getMethod());
      if (call1.getArgs() == null || call2.getArgs() == null)
      {
         assertNull("Both args should be null", call1.getArgs());
         assertNull("Both args should be null", call2.getArgs());
      }
      else
      {
         Object[] call1Args = call1.getArgs();
         Object[] call2Args = call2.getArgs();

         assertObjectArraysAreEqual(call1Args, call2Args);
      }
   }

   protected void assertObjectArraysAreEqual(Object[] a1, Object[] a2)
   {
      assertEquals("Number of args should match", a1.length, a2.length);

      for (int i = 0; i < a1.length; i++)
      {
         if (a1[i] instanceof MethodCall && a2[i] instanceof MethodCall)
         {
            assertMethodCallsEquals((MethodCall) a1[i], (MethodCall) a2[i]);
         }
         else if (a1[i] instanceof List && a2[i] instanceof List)
         {
            Object[] a1Elements = ((List) a1[i]).toArray();
            Object[] a2Elements = ((List) a2[i]).toArray();

            assertObjectArraysAreEqual(a1Elements, a2Elements);
         }
         else
         {
            assertEquals("Argument # " + i + " should be equal", a1[i], a2[i]);
         }
      }

   }


   public void testGetMarshaller()
   {
      assertEquals("Only one marshaller should be in the map by this stage", 1, marshaller.marshallers.size());

      assertEquals(expectedMarshallerClass, marshaller.getMarshaller(currentVersionShort).getClass());

      // defaultMarshaller is used for outgoing streams
      assert marshaller.defaultMarshaller.getClass().equals(expectedMarshallerClass);

      assertEquals(latestMarshallerClass, marshaller.getMarshaller(15).getClass());
      assertEquals(latestMarshallerClass, marshaller.getMarshaller(1).getClass());
      assertEquals(latestMarshallerClass, marshaller.getMarshaller(-1).getClass());
      assertEquals(latestMarshallerClass, marshaller.getMarshaller(0).getClass());
      assertEquals(CacheMarshaller200.class, marshaller.getMarshaller(20).getClass());

      assert marshaller.marshallers.size() == 2 : "Should have 2 marshallers now";
   }

   public void testStringBasedFqn() throws Exception
   {
      Fqn fqn = new Fqn<Object>("JSESSIONID", "1010.10.5:3000", "1234567890", "1");
      byte[] asBytes = marshaller.objectToByteBuffer(fqn);
      System.out.println("Marshalled to " + asBytes.length + " bytes");
      Object o2 = marshaller.objectFromByteBuffer(asBytes);
      assertEquals(fqn, o2);
   }

   public void testNonStringBasedFqn() throws Exception
   {
      Fqn fqn = new Fqn<Object>(3, false);
      byte[] asBytes = marshaller.objectToByteBuffer(fqn);
      Object o2 = marshaller.objectFromByteBuffer(asBytes);
      assertEquals(fqn, o2);
   }

   public void testMethodCall() throws Exception
   {
      Fqn fqn = new Fqn<Object>(3, false);
      MethodCall call = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal_id, fqn, "key", "value", true);
      byte[] asBytes = marshaller.objectToByteBuffer(call);
      Object o2 = marshaller.objectFromByteBuffer(asBytes);

      assertTrue("Unmarshalled object should be a method call", o2 instanceof MethodCall);
      MethodCall m2 = (MethodCall) o2;

      assertMethodCallsEquals(call, m2);
   }

   public void testNestedMethodCall() throws Exception
   {
      Fqn fqn = new Fqn<Object>(3, false);
      MethodCall call = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal_id, fqn, "key", "value", true);
      MethodCall replicateCall = MethodCallFactory.create(MethodDeclarations.replicateMethod_id, call);
      byte[] asBytes = marshaller.objectToByteBuffer(replicateCall);
      Object o2 = marshaller.objectFromByteBuffer(asBytes);
      assertTrue("Unmarshalled object should be a method call", o2 instanceof MethodCall);
      MethodCall m2 = (MethodCall) o2;

      assertMethodCallsEquals(replicateCall, m2);
   }

   public void testLargeString() throws Exception
   {
      doLargeStringTest(32767, false);
   }

   public void testLargerString() throws Exception
   {
      doLargeStringTest(32768, false);
   }

   public void test64KString() throws Exception
   {
      doLargeStringTest((2 << 15) - 10, false);
      doLargeStringTest((2 << 15) + 10, false);
   }

   public void test128KString() throws Exception
   {
      doLargeStringTest((2 << 16) - 10, false);
      doLargeStringTest((2 << 16) + 10, false);
   }

   public void testLargeStringMultiByte() throws Exception
   {
      doLargeStringTest(32767, true);
   }

   public void testLargerStringMultiByte() throws Exception
   {
      doLargeStringTest(32768, true);
   }

   public void test64KStringMultiByte() throws Exception
   {
      doLargeStringTest((2 << 15) - 10, true);
      doLargeStringTest((2 << 15) + 10, true);
   }

   public void test128KStringMultiByte() throws Exception
   {
      doLargeStringTest((2 << 16) - 10, true);
      doLargeStringTest((2 << 16) + 10, true);
   }

   protected void doLargeStringTest(int stringSize, boolean multiByteChars) throws Exception
   {
      StringBuilder sb = new StringBuilder();

      int startingChar = multiByteChars ? 210 : 65;
      for (int i = 0; i < stringSize; i++) sb.append((char) (startingChar + (i % 26)));

      String largeString = sb.toString();

      assertEquals(stringSize, largeString.length());

      byte[] buf = marshaller.objectToByteBuffer(largeString);

      assertEquals(largeString, marshaller.objectFromByteBuffer(buf));
   }

   public void testReplicationQueue() throws Exception
   {
      doReplicationQueueTest();
   }

   public void testReplicationQueueWithRegionBasedMarshalling() throws Exception
   {
      c.setUseRegionBasedMarshalling(true);
      marshaller.init();
      doReplicationQueueTest();
   }

   protected void doReplicationQueueTest() throws Exception
   {
      // replication queue takes a list of replicate() MethodCalls and wraps them in a single replicate call.
      List<MethodCall> calls = new ArrayList<MethodCall>();

      Fqn f = new Fqn<Object>("BlahBlah", 3, false);
      String k = "key", v = "value";

      MethodCall actualCall = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal_id, null, f, k, v, true);
      MethodCall replicateCall = MethodCallFactory.create(MethodDeclarations.replicateMethod_id, actualCall);

      calls.add(replicateCall);

      actualCall = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal_id, null, f, k, v, true);
      replicateCall = MethodCallFactory.create(MethodDeclarations.replicateMethod_id, actualCall);

      calls.add(replicateCall);

      MethodCall call = MethodCallFactory.create(MethodDeclarations.replicateAllMethod_id, calls);

      byte[] buf = marshaller.objectToByteBuffer(call);

      assertMethodCallsEquals(call, (MethodCall) marshaller.objectFromByteBuffer(buf));
   }

}
