package org.jboss.cache.marshall;

import org.jboss.cache.Cache;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Region;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.lock.IsolationLevel;
import static org.testng.AssertJUnit.*;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Unit test to cover replication and marshalling of custom collections
 *
 * @author <a href="mailto:galder.zamarreno@jboss.com">Galder Zamarreno</a>
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
@Test(groups = {"functional"})
public class CustomCollectionTest
{
   private Cache<Object, Object> cache1 = null;
   private Cache<Object, Object> cache2 = null;
   private String myListClass = MyList.class.getName();
   private String mySetClass = MySet.class.getName();
   private String myMapClass = MyMap.class.getName();

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      cache1 = createCache();
      cache2 = createCache();
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      cache1.stop();
      cache2.stop();
   }

   public void testMap() throws Exception
   {
      doMapTest(false);
   }

   public void testMapWithRegions() throws Exception
   {
      doMapTest(true);
   }

   public void testSet() throws Exception
   {
      doSetTest(false);
   }

   public void testSetWithRegions() throws Exception
   {
      doSetTest(true);
   }

   public void testList() throws Exception
   {
      doListTest(false);
   }

   public void testListWithRegions() throws Exception
   {
      doListTest(true);
   }

   @SuppressWarnings("unchecked")
   private void doMapTest(boolean contextClassLoader) throws Exception
   {
      ClassLoader customClassLoader = getClassLoader();
      Class mapClass = customClassLoader.loadClass(myMapClass);
      Map map = (Map) (contextClassLoader ? mapClass.newInstance() : new MyMap());

      map.put("k", "v");

      if (contextClassLoader)
      {
         enableRegionBasedClassLoading(customClassLoader);
      }

      cache1.start();
      cache2.start();

      cache1.put(fqn("/a"), "key", map);
      Object o = cache2.get(fqn("/a"), "key");
      assertTrue(o instanceof Map);
      if (contextClassLoader)
      {
         assertNotSame(MyMap.class, o.getClass());
         assertSame(mapClass, o.getClass());
      }
      else
      {
         assertSame(MyMap.class, o.getClass());
         assertNotSame(mapClass, o.getClass());
      }
      assertEquals("v", ((Map) o).get("k"));
   }

   @SuppressWarnings("unchecked")
   private void doSetTest(boolean contextClassLoader) throws Exception
   {

      ClassLoader customClassLoader = getClassLoader();
      Class setClass = customClassLoader.loadClass(mySetClass);
      Set set = (Set) (contextClassLoader ? setClass.newInstance() : new MySet());

      set.add("k");

      if (contextClassLoader)
      {
         enableRegionBasedClassLoading(customClassLoader);
      }

      cache1.start();
      cache2.start();

      cache1.put(fqn("/a"), "key", set);
      Object o = cache2.get(fqn("/a"), "key");
      assertTrue(o instanceof Set);
      if (contextClassLoader)
      {
         assertNotSame(MySet.class, o.getClass());
         assertSame(setClass, o.getClass());
      }
      else
      {
         assertSame(MySet.class, o.getClass());
         assertNotSame(setClass, o.getClass());
      }
      assertTrue(((Set) o).contains("k"));
   }

   @SuppressWarnings("unchecked")
   private void doListTest(boolean contextClassLoader) throws Exception
   {

      ClassLoader customClassLoader = getClassLoader();
      Class listClass = customClassLoader.loadClass(myListClass);

      List list = (List) (contextClassLoader ? listClass.newInstance() : new MyList());

      list.add("k");

      if (contextClassLoader)
      {
         enableRegionBasedClassLoading(customClassLoader);
      }

      cache1.start();
      cache2.start();

      cache1.put(fqn("/a"), "key", list);
      Object o = cache2.get(fqn("/a"), "key");
      assertTrue(o instanceof List);
      if (contextClassLoader)
      {
         assertSame(listClass, o.getClass());
         assertNotSame(MyList.class, o.getClass());
      }
      else
      {
         assertSame(MyList.class, o.getClass());
         assertNotSame(listClass, o.getClass());
      }
      assertTrue(((List) o).contains("k"));
   }

   public void testHarnessClassLoader() throws Exception
   {
      Class myListFromCustomLoader = getClassLoader().loadClass(myListClass);
      assertNotNull(myListFromCustomLoader);
      assertFalse(MyList.class.equals(myListFromCustomLoader));

      Object customLoaderMyList = myListFromCustomLoader.newInstance();
      try
      {
         @SuppressWarnings("unused")
         MyList m = (MyList) customLoaderMyList;
         fail("Should have barfed");
      }
      catch (ClassCastException cce)
      {
         // expected
      }
   }

   private void enableRegionBasedClassLoading(ClassLoader customClassLoader)
   {
      cache1.getConfiguration().setUseRegionBasedMarshalling(true);
      Region region1 = cache1.getRegion(fqn("/a"), true);
      region1.registerContextClassLoader(customClassLoader);
//      cache1.getRegionManager().createRegion("/a", customClassLoader);
//      cache1.getMarshaller().defaultMarshaller.registerClassLoader("/a", customClassLoader);
//      cache1.getMarshaller().defaultMarshaller.useRegionBasedMarshalling = true;
      cache2.getConfiguration().setUseRegionBasedMarshalling(true);
      Region region2 = cache2.getRegion(fqn("/a"), true);
      region2.registerContextClassLoader(customClassLoader);
//      cache2.getRegionManager().createRegion("/a", customClassLoader);
//      cache2.getMarshaller().defaultMarshaller.registerClassLoader("/a", customClassLoader);
//      cache2.getMarshaller().defaultMarshaller.useRegionBasedMarshalling = true;
   }

   private ClassLoader getClassLoader()
   {
      String[] includesClasses = {myListClass, mySetClass, myMapClass};
      String[] excludesClasses = {};
      ClassLoader cl = Thread.currentThread().getContextClassLoader();
      return new SelectedClassnameClassLoader(includesClasses, excludesClasses, cl);
   }

   private Cache<Object, Object> createCache()
   {
      Cache<Object, Object> cache = new DefaultCacheFactory().createCache(false);
      cache.getConfiguration().setCacheMode(Configuration.CacheMode.REPL_SYNC);
      cache.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache.getConfiguration().setIsolationLevel(IsolationLevel.REPEATABLE_READ);
      return cache;
   }

   private static Fqn fqn(String fqn)
   {
      return Fqn.fromString(fqn);
   }
}